/* $Id$ */
/*
 * functions to make X11 BDF font italic.
 */
#include "common.h"
#include "italic.h"
#include "utils.h"

#define PATTERN_MAX 35 /* Number of table for pixel correction + 1 */

struct pattern {
  int height0;
  int offset0;
  char *bits0[6];
  int height;
  int offset;
  char *bits[4];
};

struct bitmap {
  char **lines;
  size_t height;
  char **tmp;
};

/* correct >= 3 */
static Pattern ptable_3_1[] = {
  { 5, -2, {"..@", ".@", ".@", "@..", "@."},             1, -1, {".$"}},
  { 5, -4, {"...", "@@@.", "..@.", ".@.", "@."},         2, -2, {".$,", "$,"}},
  { 5, -2, {".....@", "....@", "...@.", "..@.", "@@."},  1,  0, {"...@$"}},
  { 4, -3, {"...@", "...@", "..@.", "@@.."},             1, -1, {".$@" }},
  { 4, -3, {"..@", "..@", ".@.", "@.."},                 1,  0, {"@$"}},
  { 4, -1, {".@", ".@", "@..", "@."},                    2, -1, {"$,", ",$"}},
  { 4, -2, {"..@", "..@@", "@@..", ".@."},               2, -1, {".$@", "@@$"}},
  { 4, -2, {".@", ".@.", "@..", "@."},                   2, -1, {"$,", ",$"}},
  { 0, 0, {NULL}, 0, 0, {NULL}}
};
/* correct >= 2 */
static Pattern ptable_2_1[] = {
  { 3, -2, {"....@.", "..@@.", "@@..."},                 1,  0, {"@@$..."}},
  { 3, -1, {"..@", ".@", ".@."},                         1,  0, {".,$"}},
  { 3, -2, {".@", ".@", "@.@"},                          1, -1, {"$@"}},
  { 3, -2, {".@", ".@.", "@.."},                         1, -1, {"$."}},
  { 3, -2, {".@", ".@@", "@.."},                         1, -1, {"$@"}},
  { 3, -1, {"..@", ".@.", "@@."},                        1,  0, {".,$"}},
  { 3, -2, {"@@.", ".@.", "@.."},                        1, -1, {"$,"}},
  { 3, -1, {"@.@", ".@.", "..."},                        1,  0, {".@$"}},
  { 3, -1, {"..@", ".@.", ".@@"},                        1,  0, {".,$"}},
  { 0, 0, {NULL}, 0, 0, {NULL}}
};
/* correct >= 3 */
static Pattern ptable_3_2[] = {
  { 6, -2, {"..@.", "...@.", "...@.", "...@.", "..@.", "@@."}, 4, -1, {"..$,", "...@", "..$,", ".$,"}},
  { 4, -1, {"@..", "@.@", ".@.", ".@."},                       1,  1, {"$@"}},
  { 4, -3, {"@.", "@@@", ".@", ".@"},                          1, -1, {"$."}},
  { 4, -3, {"..", "@@.", "..@.", "..@."},                      2, -2, {"@,", ".$,"}},
  { 3, -2, {"@..", ".@", ".@"},                                1, -1, {"$."}},
  { 4, -1, {".@.", ".@.", "..@.", "..@."},                     1,  0, {".,$"}},
  { 4, -1, {".@.", "@@.", "..@.", "..@."},                     2, -1, {"@,", "@@$"}},
  { 0, 0, {NULL}, 0, 0, {NULL}}
};
/* correct >= 2 */
static Pattern ptable_2_2[] = {
  { 3, -1, {"@..", ".@@", ".."},         1,  0, {".,@"}},
  { 3, -2, {"@..", "@@.", ".@"},         1, -1, {"$."}},
  { 3, -2, {"@.@", ".@", ".@"},          1, -1, {"$@"}},
  { 3, -1, {".@.", ".@.", "..@."},       1,  0, {"..$"}},
  { 4, -1, {"@.", "@.", ".@@", ".."},    2,  0, {",$", ".,$"}},
  { 3, -1, {"@.", "@.", ".@@"},          1,  0, {",$"}},
  { 3, -2, {"..", "@@.", "..@"},         1, -1, {"@,."}},
  { 0, 0, {NULL}, 0, 0, {NULL}}
};
/* least pattern */
static Pattern ptable_least[] = {
  { 4, -2, {"..@", ".@", "@@@", ".."},   2, -1, {"$@", "@,@"}},
  { 3, -1, {".@..", "@.@.", ".@@."},     1, 0,  {"$"}},
  { 2, -1, {".@", "@."},                 1, -1, {"$"}},
  { 0, 0, {NULL}, 0, 0, {NULL}}
};

static size_t pattern_add(Pattern *ptable, size_t i, Pattern *p);
static void pattern_print(FILE *fp, Pattern *ptable);
static char *bdf2bit(char *dst, const char *src);
static char *bit2bdf(char *dst, const char *src);

/*
 * Initialize pattern table.
 */
Pattern *pattern_new(int correct)
{
  Pattern *ptable;
  size_t i;

  ptable = XMALLOC(Pattern, PATTERN_MAX);
  i = 0;
  if (correct >= 3) {
    i += pattern_add(ptable, i, ptable_3_1);
  }
  if (correct >= 2) {
    i += pattern_add(ptable, i, ptable_2_1);
  }
  if (correct >= 3) {
    i += pattern_add(ptable, i, ptable_3_2);
  }
  if (correct >= 2) {
    i += pattern_add(ptable, i, ptable_2_2);
  }
  i += pattern_add(ptable, i, ptable_least);
  ptable[i].height0 = 0; /* sentinel */

  return ptable;
}

/*
 * Initialize pattern table.
 */
void pattern_delete(Pattern *p)
{
  XFREE(p);
}

static size_t pattern_add(Pattern *ptable, size_t i, Pattern *p)
{
  size_t count;

  count = 0;
  while (p->height0 != 0) {
    ptable[i++] = *(p++);
    count++;
  }
  return count;
}

static void pattern_print(FILE *fp, Pattern *ptable)
{
  Pattern *p;
  int i;

  p = ptable;
  while (p->height0 != 0) {
    fprintf(fp, "(%d, %d, ", p->height0, p->offset0);
    for (i = 0; i < p->height0; i++) {
      fprintf(fp, ", '%s'", p->bits0[i]);
    }
    fprintf(fp, " %d, %d, ", p->height, p->offset);
    for (i = 0; i < p->height; i++) {
      fprintf(fp, ", '%s'", p->bits[i]);
    }
    fprintf(fp, ")\n");
    p++;
  }
}

/*
 * output BITMAP image.
 */
void bitmap_output(FILE *fp, Bitmap *bitmap, int height)
{
  char **lines;
  int y;

#if !defined DEBUG
  lines = bitmap_make_bdf_tmp(bitmap, height);
#else
  lines = bitmap->lines;
#endif /* DEBUG */
  for (y = 0; y < height; y++) {
    fprintf(fp, "%s\n", lines[y]);
  }
}

/*
 * make BDF strings.
 */
char **bitmap_make_bdf_tmp(Bitmap *bitmap, int height)
{
  char **lines, **tmplines;
  char *tmp;
  int y;

  lines = bitmap->lines;
  tmplines = bitmap->tmp;
  for (y = 0; y < height; y++) {
    tmp = tmplines[y];
    bit2bdf(tmp, lines[y]);
  }
  return tmplines;
}

/*
 * Correct the pixels to make pettern more clear after the slanting.
 * Reference the ptable at the BEGIN routine.
 */
void bitmap_correct_pixel(Bitmap *bitmap, Pattern *ptable, int width, int height, int dy)
{
  Pattern *p;
  char **lines, **tmpline;
  int x, y, j, n, d, xx, nn, dd;
  char *px, *pxx, *t;

  /* add padding pixels from both side */
  lines = bitmap->lines;
  tmpline = bitmap->tmp;
  for (y = 0; y < height; y++) {
    sprintf(tmpline[y], ".%s.", lines[y]);
  }
  rstr(tmpline[height], '.', width+2);
  /* pattern matching with ptable */
  for (y = dy; y < height; y+=dy) {
    p = ptable;
    while (p->height0 != 0) {
      n = p->offset0;
      d = p->height0;
      if (((y+n) >= 0) && ((y+n+d) <= (height+1))) {
        x = 0;
        while (x < width-1) {
          px = strstr(tmpline[y+n]+x, p->bits0[0]);
          if (px == NULL) {
            break;
          }
          x = px-tmpline[y+n];
          for (j = 1; j < d; j++) {
            pxx = strstr(tmpline[y+n+j]+x, p->bits0[j]);
            if (pxx == NULL) {
              xx = -1;
              break;
            }
            xx = pxx-tmpline[y+n+j];
            if (x != xx) {
              break;
            }
          }
          if (x == xx) { /* matched ! */
            DEBUG_PRINT("==== MATCH with \"%d %d %s\" ==== (%d, %d)\n", p->height0, p->offset0, p->bits0[0], x, y);
            nn = p->offset;
            dd = p->height;
            for (j = 0; j < dd; j++) {
              t = p->bits[j];
              strncpy(tmpline[y+nn+j]+x, t, strlen(t));
            }
            break;
          }
          x++;
        }
      }
      p++;
    }
  }
  /* delete padding pixels from both side */
  for (y = 0; y < height; y++) {
    strncpy(lines[y], tmpline[y]+1, width);
    lines[y][width] = '\0';
  }
}

/*
 * Just slant the pattern of font.
 * length = dx-1+width+padding.
 */
void bitmap_make_slant(Bitmap *bitmap, int width, int height, int dx, int dy, int padding)
{
  char **lines, **tmplines;
  char *tmp;
  int y, len;

  lines = bitmap->lines;
  tmplines = bitmap->tmp;
  for (y = 0; y < height; y++) {
    tmp = tmplines[y];
    strncpy(tmp, lines[y], width);
    tmp[width] = '\0';
  }
  len = dx-1+width+padding;
  for (y = 0; y < height; y++) {
    if (y%dy == 0) {
      dx--;
    }
    rstr(lines[y], '#', len);
    strncpy(lines[y]+dx, tmplines[y], width);
  }
}

/*
 * read BITMAP and convert into image.
 */
void bitmap_get_line(Bitmap *bitmap, size_t n, const char *src)
{
  char **lines;

  lines = bitmap->lines;
  bdf2bit(lines[n], src);
}

Bitmap *bitmap_new(size_t height, size_t width)
{
  Bitmap *bitmap;
  char **p;
  size_t i;

  bitmap = XMALLOC(Bitmap, 1);
  bitmap->height = height;
  bitmap->lines = XMALLOC(char *, height);
  p = bitmap->lines;
  for (i = 0; i < height; i++) {
    *(p++) = XMALLOC(char, width+1);
  }
  bitmap->tmp = XMALLOC(char *, height+1);
  p = bitmap->tmp;
  for (i = 0; i < height+1; i++) {
    *(p++) = XMALLOC(char, width+1);
  }
  return bitmap;
}

void bitmap_delete(Bitmap *bitmap)
{
  char **p;
  size_t i, height;

  height = bitmap->height;
  p = bitmap->lines;
  for (i = 0; i < height; i++) {
    XFREE(*p);
    p++;
  }
  XFREE(bitmap->lines);
  p = bitmap->tmp;
  for (i = 0; i < height+1; i++) {
    XFREE(*p);
    p++;
  }
  XFREE(bitmap->tmp);
  XFREE(bitmap);
}

/*
 * recast BBX or FONTBOUNDINGBOX.
 */
void recast_bbx(char *dst, int *padding, int *width, int *height, int *dx, int dy, const char *src)
{
  char tmp[16];
  int w, h, xd, yd;

  sscanf(src, "%15s%d%d%d%d", tmp, &w, &h, &xd, &yd);
  *width = w;
  *height = h;
  *dx = (h+dy-1)/dy;
  if ((h > 0) && (w > 0)) {
    w += *dx-1;
    xd -= (*dx-1)/2;
    *padding = (8-w%8)%8;
  } else {
    *padding = 0;
  }
  sprintf(dst, "%s %d %d %d %d\n", tmp, w, h, xd, yd);
}

/*
 * convert BDF format into bit image.
 */
static char *bdf2bit(char *dst, const char *src)
{
  const char *s, *p;
  char *d;
  unsigned int c, b;

  d = dst;
  s = src;
  while (sscanf(s, "%02x", &c) == 1) {
    b = 0x80;
    while (b > 0x00) {
      p = (c&b) ? "@" : ".";
      *(d++) = *p;
      b >>= 1;
    }
    s+=2;
  }
  *d = '\0';
  return dst;
}

/*
 * convert bit image into BDF format.
 */
static char *bit2bdf(char *dst, const char *src)
{
  const char *s;
  char *d;
  char tmp[9];
  unsigned int b;
  int i;

  s = src;
  d = dst;
  while (sscanf(s, "%8[@$.#,]", tmp) == 1) {
    b = 0x00;
    for (i = 0; i < 8; i++) {
      switch (tmp[i]) {
      case '@':
      case '$':
        b |= (0x01 << (7-i));
        break;
      case '.':
      case '#':
      case ',':
      default:
        break;
      }
    }
    s += 8;
    sprintf(d, "%02X", b);
    d += 2;
  }
  return dst;
}
